/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmQaManager.h
/// \brief  Manager task for other QA taska (implementation)
/// \author Sergei Zharko <s.zharko@gsi.de>
/// \since  17.10.2023

#include "CbmQaManager.h"

#include "CbmQaTask.h"
#include "Logger.h"

#include <iomanip>

// ---------------------------------------------------------------------------------------------------------------------
//
CbmQaManager::CbmQaManager(int verbose) : FairTask("CbmQaManager", verbose) {}


// ---------------------------------------------------------------------------------------------------------------------
//
void CbmQaManager::Finish()
{
  using std::left;
  using std::right;
  using std::setw;

  for (auto* task : *(this->GetListOfTasks())) {
    auto* pQaTask = dynamic_cast<CbmQaTask*>(task);
    if (pQaTask) {
      LOG(info) << "Checking the task " << pQaTask->GetName();
      pQaTask->Check();
      pQaTask->ReadCheckListFromConfig();
    }
  }

  LOG(info) << fName << ": monitorable status:";
  fStatus = true;
  for (auto* task : *(this->GetListOfTasks())) {
    auto* pQaTask = dynamic_cast<CbmQaTask*>(task);
    if (pQaTask) {
      LOG(info) << "Check list for the task " << pQaTask->GetName();
      const auto& mCheckList = pQaTask->GetCheckList();
      for (const auto& [entryName, flags] : mCheckList) {
        LOG(info) << '\t' << left << setw(40) << entryName << right << setw(10)
                  << (flags.fResult ? "\e[1;32mpassed\e[0m" : "\e[1;31mfailed\e[0m")
                  << (flags.fStatus ? "" : " IGNORED");
        if (flags.fStatus) {
          fStatus &= flags.fResult;
        }
      }
    }
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmQaManager::Init()
{
  // Apply configuration file to all underlying tasks
  if (fsConfigName.Length()) {
    LOG(info) << "CbmQaManager: using configuration file " << fsConfigName;
    for (auto* task : *(this->GetListOfTasks())) {
      auto* pQaTask = dynamic_cast<CbmQaTask*>(task);
      if (pQaTask) {
        pQaTask->SetConfigName(fsConfigName.Data());
      }
    }
  }
  return kSUCCESS;
}
