/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */


#include "CaToolsMaterialHelper.h"

#include "Logger.h"
#include "TGeoManager.h"
#include "TGeoMedium.h"
#include "TGeoNavigator.h"
#include "TGeoNode.h"
#include "TGeoPhysicalNode.h"
#include "TGeoVoxelFinder.h"

#include <iostream>
#include <thread>

using cbm::ca::tools::MaterialHelper;

using cbm::algo::ca::MaterialMap;

ClassImp(MaterialHelper);

MaterialHelper::MaterialHelper()
{
  // constructor

  // save the current number of threads to restore it when the work is done
  fNthreadsOld = gGeoManager->GetMaxThreads();

  // set defaut n threads to 1

  fNthreads = 1;

  LOG(info) << "------------------------------------------------------------------------------------";
  LOG(info) << "                      create L1 material maps ";
  LOG(info) << "";
  LOG(info) << "  !! To run it with the full speed, set: \"export OMP_NUM_THREADS=<n CPUs>\" before running !!";
  LOG(info) << "";
  // if possible, read the allowed n threads from the environment variable
  {
    const char* env = std::getenv("OMP_NUM_THREADS");
    if (env) {
      fNthreads = TString(env).Atoi();
      LOG(info) << "  found environment variable OMP_NUM_THREADS = \"" << env << "\", read as integer: " << fNthreads;
    }
  }

  // ensure that at least one thread is set

  if (fNthreads < 1) {
    fNthreads = 1;
  }

  LOG(info) << "  Maps will be created using " << fNthreads << " CPU threads";
  LOG(info) << "";
  LOG(info) << "  It might crash because of a ROOT bug. In this case do one of the following: ";
  LOG(info) << "";
  LOG(info) << "    - fix illegal overlaps in the geometry via gGeoManager->CheckOverlaps()";
  LOG(info) << "    - run with CbmL1::SetSafeMaterialInitialization() option. ";
  LOG(info) << "      It fixes the crash, but slows down the initialization significantly.";
  LOG(info) << "    - manually disable voxel finders for problematic volumes in CaToolsMaterialHelper.cxx";
  LOG(info) << "";
  LOG(info) << "------------------------------------------------------------------------------------";

  InitThreads();

  //
  // It looks like TGeoVoxelFinder contains a bug and therefore sometimes produces a crash.
  //
  // Currently, the crash only happens on some of TGeoVolumeAssembly volumes
  // and only when an (inproper?) alignment matrices are applied.
  //
  // Here we try to catch this case and disable those finders.
  //
  // Disabling all the voxel finders leads to a very long initialization, try to avoid it!
  //
  // TODO: make a bug report to the ROOT team; remove the following code once the bug is fixed.
  //

  if (fDoSafeInitialization) {  // disable all voxel finders
    TObjArray* volumes = gGeoManager->GetListOfVolumes();
    if (volumes) {
      for (int iv = 0; iv < volumes->GetEntriesFast(); iv++) {
        TGeoVolume* vol = dynamic_cast<TGeoVolume*>(volumes->At(iv));
        if (!vol) {
          continue;
        }
        TGeoVoxelFinder* finder = vol->GetVoxels();
        if (finder) {
          finder->SetInvalid();
        }
      }
    }
  }
  else {  // disable some voxel finders in some cases

    // check if any alignment was applied

    bool isAlignmentApplied = false;
    {
      TObjArray* nodes = gGeoManager->GetListOfPhysicalNodes();
      if (nodes) {
        for (int in = 0; in < nodes->GetEntriesFast(); in++) {
          TGeoPhysicalNode* node = dynamic_cast<TGeoPhysicalNode*>(nodes->At(in));
          if (!node) continue;
          if (node->IsAligned()) {
            isAlignmentApplied = true;
            break;
          }
        }
      }
    }

    // disable potentially problematic voxel finders

    if (isAlignmentApplied) {
      TObjArray* volumes = gGeoManager->GetListOfVolumes();
      if (volumes) {
        for (int iv = 0; iv < volumes->GetEntriesFast(); iv++) {
          TGeoVolumeAssembly* vol = dynamic_cast<TGeoVolumeAssembly*>(volumes->At(iv));
          if (!vol) {
            continue;
          }
          TGeoVoxelFinder* finder = vol->GetVoxels();
          if (finder) {
            finder->SetInvalid();
          }
        }
      }
    }
  }
}


MaterialHelper::~MaterialHelper()
{
  // destructor

  CleanUpThreads();

  // delete the navigators, created in this class

  for (auto i = fNavigators.begin(); i != fNavigators.end(); i++) {
    gGeoManager->RemoveNavigator(*i);
  }

  // once TGeoManager is swithched in multithreaded mode, there is no way to make it non-mltithreaded again
  // therefore we should set SetMaxThreads( >=0 )

  if (fNthreadsOld > 0) {
    gGeoManager->SetMaxThreads(fNthreadsOld);
  }
  else {
    gGeoManager->SetMaxThreads(1);
  }

  // ensure that the default navigator exists

  GetCurrentNavigator(-1);
}


TGeoNavigator* MaterialHelper::GetCurrentNavigator(int iThread)
{
  // Get navigator for current thread, create it if it doesnt exist.
  // Produce an error when anything goes wrong

  TGeoNavigator* navi = gGeoManager->GetCurrentNavigator();
  if (!navi) {

    navi = gGeoManager->AddNavigator();

    if (iThread >= 0) {
      fNavigators.push_back(navi);
    }

    if (navi != gGeoManager->GetCurrentNavigator()) {
      LOG(fatal) << "ca::tools::MaterialHelper: unexpected behavior of TGeoManager::AddNavigator() !!";
    }
  }

  if (!navi) {
    LOG(fatal) << "ca::tools::MaterialHelper: can not find / create TGeoNavigator for thread " << iThread;
  }

  return navi;
}


void MaterialHelper::CleanUpThreads()
{
  // clean up thread data in TGeoManager

  gGeoManager->ClearThreadsMap();

  // create a default navigator for multithreaded mode
  // otherwise gGeoManager->GetCurrentNavigator() returns nullptr that might produces crashes in other code

  GetCurrentNavigator(-1);
}

void MaterialHelper::InitThreads()
{
  // (re)initialise the number of threads in TGeoManager

  // reserve enough threads in TGeoManager
  if (fNthreads > gGeoManager->GetMaxThreads()) {
    gGeoManager->SetMaxThreads(fNthreads);
    // in case the number of threads is truncated by TGeoManager (must not happen)
    fNthreads = gGeoManager->GetMaxThreads();
    if (fNthreads < 1) {
      fNthreads = 1;
    }
  }
  CleanUpThreads();
}


MaterialMap MaterialHelper::GenerateMaterialMap(double zRef, double zMin, double zMax, double xyMax, int nBinsDim)
{
  if (fDoRadialProjection) {
    if (fTargetZ >= zRef - 1.) {
      LOG(fatal) << "ca::tools::MaterialHelper: the material reference Z must be at least 1 cm downstream the target: "
                    "target Z = "
                 << fTargetZ << ", material reference z = " << zRef;
    }
    if (zMin < fTargetZ) {
      zMin = fTargetZ;
    }
  }

  if (!(zMin <= zRef && zRef <= zMax)) {
    LOG(fatal) << "ca::tools::MaterialHelper: the material parameters are inconsistent. It must be: zMin " << zMin
               << " <= zRef " << zRef << " <= zMax " << zMax;
  }

  MaterialMap matBudget;
  matBudget.Initialize(nBinsDim, xyMax, zRef, zMin, zMax);
  double binSizeX = 2. * xyMax / nBinsDim;
  double binSizeY = 2. * xyMax / nBinsDim;

  // call it every time. for the case the number of threads was reset in some other code
  InitThreads();

  long nThreadCrosses[fNthreads];
  long nThreadRays[fNthreads];
  long nThreadRaysExpected = nBinsDim * nBinsDim * fNraysBinPerDim * fNraysBinPerDim / fNthreads;
  for (int i = 0; i < fNthreads; i++) {
    nThreadCrosses[i] = 0;
    nThreadRays[i]    = 0;
  }

  LOG(info) << "ca::tools::MaterialHelper: Generate material map using " << fNthreads << " threads..";

  auto naviThread = [&](int iThread) {
    // create a navigator for this thread
    TGeoNavigator* navi = GetCurrentNavigator(iThread);

    bool doPrint = false;  // flag for debug printing

    for (int iBinX = iThread; iBinX < nBinsDim; iBinX += fNthreads) {
      for (int iBinY = 0; iBinY < nBinsDim; ++iBinY) {
        double radThick = 0.;
        int nRays       = 0;
        double d        = 1. / (fNraysBinPerDim);
        for (int iStepX = 0; iStepX < fNraysBinPerDim; iStepX++) {
          for (int iStepY = 0; iStepY < fNraysBinPerDim; iStepY++) {

            // ray position at zRef
            double rayX = -xyMax + (iBinX + d * (iStepX + 0.5)) * binSizeX;
            double rayY = -xyMax + (iBinY + d * (iStepY + 0.5)) * binSizeY;

            // ray slopes
            double tx = 0.;
            double ty = 0.;
            double t  = 1.;
            if (fDoRadialProjection) {
              tx = rayX / (zRef - fTargetZ);
              ty = rayY / (zRef - fTargetZ);
              t  = sqrt(1. + tx * tx + ty * ty);
            }

            // ray position at zMin
            double z = zMin;
            double x = rayX + tx * (z - zRef);
            double y = rayY + ty * (z - zRef);
            if (doPrint) {
              std::cout << "ray at " << x << " " << y << " " << z << " zMin " << zMin << " zMax " << zMax << std::endl;
            }

            TGeoNode* node = navi->InitTrack(x, y, z, tx / t, ty / t, 1. / t);

            bool doContinue = 1;
            for (int iStep = 0; doContinue; iStep++) {

              if (!node) {
                // may happen when tracing outside of the CBM volume -> produce a warning
                LOG(warning) << "ca::tools::MaterialHelper: TGeoNavigator can not find the geo node";
                break;
              }

              TGeoMedium* medium = node->GetMedium();
              if (!medium) {
                LOG(fatal) << "ca::tools::MaterialHelper: TGeoNavigator can not find the geo medium";
              }

              TGeoMaterial* material = medium->GetMaterial();
              if (!material) {
                LOG(fatal) << "ca::tools::MaterialHelper: TGeoNavigator can not find the geo material";
              }

              double radLen = material->GetRadLen();

              if (radLen < 0.56) {  // 0.5612 is rad. length of Lead at normal density
                LOG(fatal) << "ca::tools::MaterialHelper: failed assertion!";
                LOG(fatal) << "     TGeoNavigator returns a suspicious material with an unrealistic "
                           << "radiation length of " << radLen << " cm.";
                LOG(fatal) << "     The allowed minimum is 0.56 cm (Lead has 0.5612 cm). Check your material! "
                              "Modify this assertion if necessary.";
                LOG(fatal) << "     TGeoNode \"" << node->GetName() << "\", TGeoMedium \"" << medium->GetName()
                           << "\", TGeoMaterial \"" << material->GetName() << "\"";
              }

              // move navi to the next material border
              node = navi->FindNextBoundaryAndStep();  //5., kTRUE);
              nThreadCrosses[iThread]++;

              if (doPrint) {
                std::cout << "   RL " << radLen << " next pt " << navi->GetCurrentPoint()[0] << " "
                          << navi->GetCurrentPoint()[1] << " " << navi->GetCurrentPoint()[2] << std::endl;
              }

              double zNew = navi->GetCurrentPoint()[2];
              if (zNew >= zMax) {
                zNew       = zMax;
                doContinue = 0;
              }

              //if (zNew < z) {
              //std::cout << " z curr " << z << " z new " << zNew << " dz " << zNew - z << std::endl; }
              double dz = zNew - z;
              if (dz < 0.) {
                if (iStep == 0 && dz > -1.e-8) {
                  // when the ray starts exactly at the volume border, first dz might become negative,
                  // probably due to some roundoff errors. So we let it be a bit negative for the first intersection.
                  dz = 0.;
                }
                else {
                  LOG(fatal)
                    << "ca::tools::MaterialHelper: TGeoNavigator propagates the ray upstream. Something is wrong."
                    << " z old " << z << " z new " << zNew << ", dz " << dz;
                }
              }
              radThick += dz / radLen;
              z = zNew;
            }

            nRays++;
            nThreadRays[iThread]++;
            if (nThreadRays[iThread] % 1000000 == 0) {
              LOG(info) << "ca::tools::MaterialHelper:  report from thread " << iThread << ":   material map is "
                        << 100. * nThreadRays[iThread] / nThreadRaysExpected << " \% done";
            }
          }
        }
        radThick /= nRays;
        if (doPrint) {
          std::cout << "   radThick " << radThick << std::endl;
        }
        //doPrint = (radThick > 0.01);
        matBudget.SetRadThickBin(iBinX, iBinY, radThick);
      }  // iBinX
    }    // iBinY
  };

  std::vector<std::thread> threads(fNthreads);

  // run n threads
  for (int i = 0; i < fNthreads; i++) {
    threads[i] = std::thread(naviThread, i);
  }

  // wait for the threads to finish
  for (auto& th : threads) {
    th.join();
  }

  CleanUpThreads();

  long nRays    = 0;
  long nCrosses = 0;
  for (int i = 0; i < fNthreads; i++) {
    nRays += nThreadRays[i];
    nCrosses += nThreadCrosses[i];
  }

  LOG(info) << "ca::tools::MaterialHelper:     shooted " << nRays << " rays. Each ray crossed " << nCrosses * 1. / nRays
            << " material boundaries in average";

  return matBudget;
}
